/*
 * RaceTrap.java
 *
 * Created on 22. juni 2000, 13:48
 * 
 * Brian Vinter
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graphicsScreen.h"
#include "StopWatch.h"

typedef struct {
  double dLength;
  unsigned char iSet;
  unsigned char baPath[0];
}  cRouteDefinition; 


int iSize;                      //Number of grain-bags
int **iaaCoors;              //Coordinates for the grain-bags
double **daaDistanceTable;   //Full table for distance between any two grain-bags
double dMinRouteLen=10E100;//Large number used for initial best distance
double dGlobalBest=10E100;//Bounding variable

#ifdef GRAPHICS
void zPlotRoute(char *baOrder){ //This function simply plots a route to the display
  int i;
  gs_clear(WHITE);
  for(i=0; i<iSize; i++)
    gs_dot(iaaCoors[i][0], iaaCoors[i][1], 10, RED); 
  for(i=0; i<iSize-1; i++)
    gs_line(iaaCoors[baOrder[i]][0], iaaCoors[baOrder[i]][1], iaaCoors[baOrder[i+1]][0], iaaCoors[baOrder[i+1]][1], RED);    
  gs_line(iaaCoors[baOrder[iSize-1]][0], iaaCoors[baOrder[iSize-1]][1], iaaCoors[baOrder[0]][0], iaaCoors[baOrder[0]][1], RED);
  gs_update();
}
#endif
 

cRouteDefinition *zRoute(cRouteDefinition *oRoute){ //This is our recursive route finder
  unsigned char c;
  int i;
  double dNewLength;
  cRouteDefinition *oRouteDefinition, *oPermutation, *oResult, *oBest;
  
  oBest=(cRouteDefinition *)malloc(sizeof(cRouteDefinition)+iSize);//The subtree that will eventually be returned
  
  oBest->dLength=dMinRouteLen;
  
  if(oRoute->iSet==iSize){ //If all grain-bags have been placed we simply add the distance to get beck home
    oRoute->dLength+=daaDistanceTable[oRoute->baPath[oRoute->iSet-1]][oRoute->baPath[0]];
    return oRoute;
  }
  
  c=oRoute->baPath[oRoute->iSet];
  for(i=oRoute->iSet;i<iSize;i++){
    dNewLength=oRoute->dLength+daaDistanceTable[oRoute->baPath[oRoute->iSet-1]][oRoute->baPath[i]];
    //If the partial route is smaller that the best known route we probe deeber into the tree
    if(dNewLength<dGlobalBest){
      oPermutation=(cRouteDefinition *)malloc(sizeof(cRouteDefinition)+iSize); //A new instance of route where we create a new permutation
      memcpy(oPermutation->baPath,oRoute->baPath, iSize);   //Inherit the path
      oPermutation->baPath[oRoute->iSet]=oRoute->baPath[i];
      oPermutation->baPath[i]=c;
      oPermutation->iSet=oRoute->iSet;
      oPermutation->iSet++;   //One more bag is placed!
      //Add the new bag to the total distance for the path
      oPermutation->dLength=dNewLength;
      oResult = zRoute(oPermutation);
      if(oResult->dLength<oBest->dLength){ //Best route so far?
	free(oBest);
	oBest=oResult;
	if(oBest->dLength<dGlobalBest){
	  dGlobalBest=oBest->dLength;
#ifdef GRAPHICS
	  zPlotRoute((char *)oBest->baPath);
#endif
	}
      }
      else {
	free(oResult);
      }
    }
  }
  free(oRoute);
  
  return oBest;
}
    
double zEuclidDist(int *from, int *to){ //In the dessert the shortest route is a straight line:)
  double dx=fabs(from[0]-to[0]);
  double dy=fabs(from[1]-to[1]);
  return(sqrt(dx*dx+dy*dy));
  }
    
static void zReadRoute(){ //Reads coordinates from a file and generates a distance-table - ugly as hell!! :)

  FILE *file=fopen("./route.dat", "r");
  int i,j;

  fscanf(file,"%d",&iSize);

  daaDistanceTable=(double **)(malloc(iSize*sizeof(double *)));
  for(i=0;i<iSize;i++)
    daaDistanceTable[i]=(double *)(malloc(iSize*sizeof(double)));
  iaaCoors=(int **)(malloc(iSize*sizeof(int *)));
  for(i=0;i<iSize;i++)
    iaaCoors[i]=(int *)(malloc(2*sizeof(int)));
  
  for(i=0;i<iSize;i++)
    fscanf(file,"%d %d",&(iaaCoors[i][0]),&(iaaCoors[i][1]));

    for(i=0; i<iSize; i++)	  
      for(j=0; j<iSize; j++)	  
	daaDistanceTable[i][j]=zEuclidDist(iaaCoors[i],iaaCoors[j]);
  }
    
    
int main (int argc, unsigned char **args) {
  
  cRouteDefinition *oOriginalRoute,*res;
  int i;
  char buf[256];

  zReadRoute();
#ifdef GRAPHICS
 gs_init(501,501);
#endif

  oOriginalRoute=(cRouteDefinition *)malloc(sizeof(cRouteDefinition)+iSize);
  for(i=0; i<iSize; i++)
    oOriginalRoute->baPath[i]=(unsigned char)i;
  
#ifdef GRAPHICS
  zPlotRoute((char *)oOriginalRoute->baPath); //Show the original route
#endif
  
  oOriginalRoute->dLength=0.0;
  oOriginalRoute->iSet=1;
  
  sw_init();
  sw_start();
  res=zRoute(oOriginalRoute);  //Find the best route:)
  sw_stop();
  sw_timeString(buf);
#ifdef GRAPHICS
  zPlotRoute((char *)res->baPath);           //Show the best route
#endif  
  printf("Route length is %lf it took %s\n", res->dLength,buf);
#ifdef GRAPHICS
  gs_exit();
#endif  
}
